"""Module for handling devices like Lights, Switches or Covers."""
from .devices import Devices
from .action import Action, ActionBase
from .cover import Cover
from .travelcalculator import TravelCalculator
from .climate import Climate, ClimateOperationMode
from .light import Light
from .switch import Switch
from .time import Time
from .sensor import Sensor
from .binary_sensor import BinarySensor, BinarySensorState
from .notification import Notification



"""
Module for reading the value of a specific KNX group address from KNX bus.

The module will
* ... send a group_read to the selected gruop address.
* ... register a callback for receiving telegrams within telegram queue.
* ... check if received telegrams have the correct group address.
* ... store the received telegram for further processing.
"""

import asyncio

from xknx.knx import Telegram, TelegramType

class ValueReader:
    """Class for reading the value of a specific KNX group address from KNX bus."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, xknx, group_address, timeout_in_seconds=1):
        """Initialize ValueReader class."""
        self.xknx = xknx
        self.group_address = group_address
        self.response_received_or_timeout = asyncio.Event()
        self.success = False
        self.timeout_in_seconds = timeout_in_seconds
        self.timeout_callback = None
        self.timeout_handle = None
        self.received_telegram = None

    @asyncio.coroutine
    def read(self):
        """Send group read and wait for response."""
        @asyncio.coroutine
        def telegram_received_callback(telegram):
            """Callback for having received a telegram."""
            return self.telegram_received(telegram)

        cb_obj = self.xknx.telegram_queue.register_telegram_received_cb(
            telegram_received_callback)

        yield from self.send_group_read()
        yield from self.start_timeout()
        yield from self.response_received_or_timeout.wait()
        yield from self.stop_timeout()

        self.xknx.telegram_queue.unregister_telegram_received_cb(
            cb_obj)

        if not self.success:
            return None
        return self.received_telegram


    @asyncio.coroutine
    def send_group_read(self):
        """Send group read."""
        telegram = Telegram(self.group_address, TelegramType.GROUP_READ)
        yield from self.xknx.telegrams.put(telegram)

    def telegram_received(self, telegram):
        """Callback for having received a telegram. Check if group addresses match."""
        if self.group_address == telegram.group_address:
            self.success = True
            self.received_telegram = telegram
            self.response_received_or_timeout.set()
            return True
        return False

    def timeout(self):
        """Callback for handling timeout."""
        self.response_received_or_timeout.set()

    @asyncio.coroutine
    def start_timeout(self):
        """Start timeout. Register callback for no answer received within timeout."""
        self.timeout_handle = self.xknx.loop.call_later(
            self.timeout_in_seconds, self.timeout)

    @asyncio.coroutine
    def stop_timeout(self):
        """Stop timeout."""
        self.timeout_handle.cancel()



"""Module for the automations and business logic of XKNX."""
from .stateupdater import StateUpdater
from .telegram_queue import TelegramQueue
from .config import Config
from .value_reader import ValueReader



"""
KNXIPInterface manages KNX/IP Tunneling or Routing connections.

* It searches for available devices and connects with the corresponding connect method.
* It passes KNX telegrams from the network and
* provides callbacks after having received a telegram from the network.

"""

import asyncio
from enum import Enum
from xknx.exceptions import XKNXException
from .gateway_scanner import GatewayScanner

from .routing import Routing
from .tunnel import Tunnel
from .const import DEFAULT_MCAST_PORT

class ConnectionType(Enum):
    """Enum class for different types of KNX/IP Connections."""

    AUTOMATIC = 0
    TUNNELING = 1
    ROUTING = 2

class ConnectionConfig:
    """
    Connection configuration.

    Handles:
    * type of connection:
        * AUTOMATIC for using GatewayScanner for searching and finding KNX/IP devices in the network.
        * TUNNELING connect to a specific KNX/IP tunneling device.
        * ROUTING use KNX/IP multicast routing.
    * local_ip: Local ip of the interface though which KNXIPInterface should connect.
    * gateway_ip: IP of KNX/IP tunneling device.
    * gateway_port: Port of KNX/IP tunneling device.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self,
                 connection_type=ConnectionType.AUTOMATIC,
                 local_ip=None,
                 gateway_ip=None,
                 gateway_port=DEFAULT_MCAST_PORT):
        """Initialize ConnectionConfig class."""
        self.connection_type = connection_type
        self.local_ip = local_ip
        self.gateway_ip = gateway_ip
        self.gateway_port = gateway_port



class KNXIPInterface():
    """Class for managing KNX/IP Tunneling or Routing connections."""

    def __init__(self, xknx, connection_config=ConnectionConfig()):
        """Initialize KNXIPInterface class."""
        self.xknx = xknx
        self.interface = None
        self.connection_config = connection_config

    @asyncio.coroutine
    def start(self):
        """Start interface. Connecting KNX/IP device with the selected method."""
        if self.connection_config.connection_type == ConnectionType.AUTOMATIC:
            yield from self.start_automatic()
        elif self.connection_config.connection_type == ConnectionType.ROUTING:
            yield from self.start_routing(
                self.connection_config.local_ip)
        elif self.connection_config.connection_type == ConnectionType.TUNNELING:
            yield from self.start_tunnelling(
                self.connection_config.local_ip,
                self.connection_config.gateway_ip,
                self.connection_config.gateway_port)

    @asyncio.coroutine
    def start_automatic(self):
        """Start GatewayScanner and connect to the found device."""
        gatewayscanner = GatewayScanner(self.xknx)
        yield from gatewayscanner.start()
        gatewayscanner.stop()

        if not gatewayscanner.found:
            raise XKNXException("No Gateways found")

        if gatewayscanner.supports_tunneling:
            yield from self.start_tunnelling(gatewayscanner.found_local_ip,
                                             gatewayscanner.found_ip_addr,
                                             gatewayscanner.found_port)

        elif gatewayscanner.supports_routing:
            yield from self.start_routing(gatewayscanner.found_local_ip)


    @asyncio.coroutine
    def start_tunnelling(self, local_ip, gateway_ip, gateway_port):
        """Start KNX/IP tunnel."""
        self.xknx.logger.debug("Starting tunnel to %s:%s from %s", gateway_ip, gateway_port, local_ip)
        self.interface = Tunnel(
            self.xknx,
            self.xknx.own_address,
            local_ip=local_ip,
            gateway_ip=gateway_ip,
            gateway_port=gateway_port,
            telegram_received_callback=self.telegram_received)
        yield from self.interface.start()

    @asyncio.coroutine
    def start_routing(self, local_ip):
        """Start KNX/IP Routing."""
        self.xknx.logger.debug("Starting Routing from %s", local_ip)
        self.interface = Routing(
            self.xknx,
            self.telegram_received,
            local_ip)
        yield from self.interface.start()


    @asyncio.coroutine
    def stop(self):
        """Stop connected interfae (either Tunneling or Routing)."""
        if self.interface is not None:
            yield from self.interface.stop()
            self.interface = None

    def telegram_received(self, telegram):
        """Callback for having received telegram."""
        self.xknx.loop.create_task(
            self.xknx.telegrams.put(telegram))


    @asyncio.coroutine
    def send_telegram(self, telegram):
        """Send telegram to connected device (either Tunneling or Routing)."""
        yield from self.interface.send_telegram(telegram)



"""
Module for handling KNX primitves like:

* KNX Addresses
* KNX Values like Int, Float, Time
* Derived KNX Values like Scaling, Time, Temperature

"""
from .address import Address, AddressType, AddressFormat
from .address_filter import AddressFilter
from .telegram import Telegram, TelegramDirection, TelegramType
from .dpt import DPTBase, DPTBinary, DPTArray, DPTComparator
from .dpt_float import DPTFloat, DPTLux, DPTTemperature, DPTHumidity, DPTWsp
from .dpt_2byte import DPTUElCurrentmA
from .dpt_scaling import DPTScaling
from .dpt_time import DPTTime, DPTWeekday
from .dpt_string import DPTString



"""
GatewayScanner is an abstraction for searching for KNX/IP devices on the local network.

* It walks through all network interfaces
* and sends UDP multicast search requests
* it returns the first found device
"""

import asyncio
import netifaces
from xknx.knxip import HPAI, KNXIPFrame, SearchResponse, KNXIPServiceType, \
    DIBSuppSVCFamilies, DIBServiceFamily
from .udp_client import UDPClient
from .const import DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT

class GatewayScanner():
    """Class for searching KNX/IP devices."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, xknx, timeout_in_seconds=1):
        """Initialize GatewayScanner class."""
        self.xknx = xknx
        self.response_received_or_timeout = asyncio.Event()
        self.found = False
        self.found_ip_addr = None
        self.found_port = None
        self.found_name = None
        self.found_local_ip = None
        self.supports_routing = False
        self.supports_tunneling = False
        self.udpclients = []
        self.timeout_in_seconds = timeout_in_seconds
        self.timeout_callback = None
        self.timeout_handle = None

    def response_rec_callback(self, knxipframe, udp_client):
        """Callback for receiving a SearchResponse."""
        if not isinstance(knxipframe.body, SearchResponse):
            self.xknx.logger.warning("Cant understand knxipframe")
            return

        if not self.found:
            self.found_ip_addr = knxipframe.body.control_endpoint.ip_addr
            self.found_port = knxipframe.body.control_endpoint.port
            self.found_name = knxipframe.body.device_name

            for dib in knxipframe.body.dibs:
                if isinstance(dib, DIBSuppSVCFamilies):
                    self.supports_routing = dib.supports(DIBServiceFamily.ROUTING)
                    self.supports_tunneling = dib.supports(DIBServiceFamily.TUNNELING)

            (self.found_local_ip, _) = udp_client.getsockname()

            self.response_received_or_timeout.set()
            self.found = True


    @asyncio.coroutine
    def start(self):
        """Start searching."""
        yield from self.send_search_requests()
        yield from self.start_timeout()
        yield from self.response_received_or_timeout.wait()
        yield from self.stop()
        yield from self.stop_timeout()

    @asyncio.coroutine
    def stop(self):
        """Stop tearing down udpclient."""
        for udpclient in self.udpclients:
            yield from udpclient.stop()


    @asyncio.coroutine
    def send_search_requests(self):
        """Send search requests on all connected interfaces."""
        # pylint: disable=no-member
        for interface in netifaces.interfaces():
            try:
                af_inet = netifaces.ifaddresses(interface)[netifaces.AF_INET]
                ip_addr = af_inet[0]["addr"]
                yield from self.search_interface(interface, ip_addr)
            except KeyError:
                self.xknx.logger.info("Could not connect to an KNX/IP device on %s", interface)
                continue


    @asyncio.coroutine
    def search_interface(self, interface, ip_addr):
        """Search on a specific interface."""
        self.xknx.logger.debug("Searching on %s / %s", interface, ip_addr)

        udpclient = UDPClient(self.xknx,
                              (ip_addr, 0),
                              (DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT),
                              multicast=True)

        udpclient.register_callback(
            self.response_rec_callback, [KNXIPServiceType.SEARCH_RESPONSE])
        yield from udpclient.connect()

        self.udpclients.append(udpclient)

        (local_addr, local_port) = udpclient.getsockname()
        knxipframe = KNXIPFrame(self.xknx)
        knxipframe.init(KNXIPServiceType.SEARCH_REQUEST)
        knxipframe.body.discovery_endpoint = \
            HPAI(ip_addr=local_addr, port=local_port)
        knxipframe.normalize()

        udpclient.send(knxipframe)


    def timeout(self):
        """Handle timeout for not having received a SearchResponse."""
        self.response_received_or_timeout.set()


    @asyncio.coroutine
    def start_timeout(self):
        """Start time out."""
        self.timeout_handle = self.xknx.loop.call_later(
            self.timeout_in_seconds, self.timeout)

    @asyncio.coroutine
    def stop_timeout(self):
        """Stop/cancel timeout."""
        self.timeout_handle.cancel()



"""
XKNX is an Asynchronous Python module for reading and writing KNX/IP packets.
"""
import asyncio
import signal
import logging
from xknx.knx import Address, AddressFormat
from xknx.io import KNXIPInterface, ConnectionConfig
from xknx.core import TelegramQueue, Config
from xknx.devices import Devices

class XKNX:
    """Class for reading and writing KNX/IP packets."""

    # pylint: disable=too-many-instance-attributes

    DEFAULT_ADDRESS = '15.15.250'

    def __init__(self,
                 config=None,
                 loop=None,
                 own_address=Address(DEFAULT_ADDRESS),
                 address_format=AddressFormat.LEVEL3,
                 telegram_received_cb=None,
                 device_updated_cb=None):
        """Initialize XKNX class."""
        # pylint: disable=too-many-arguments
        self.devices = Devices()
        self.telegrams = asyncio.Queue()
        self.loop = loop or asyncio.get_event_loop()
        self.sigint_received = asyncio.Event()
        self.telegram_queue = TelegramQueue(self)
        self.state_updater = None
        self.knxip_interface = None
        self.started = False
        self.address_format = address_format
        self.own_address = own_address
        self.logger = logging.getLogger('xknx.log')
        self.knx_logger = logging.getLogger('xknx.knx')
        self.telegram_logger = logging.getLogger('xknx.telegram')

        if config is not None:
            Config(self).read(config)

        if telegram_received_cb is not None:
            self.telegram_queue.register_telegram_received_cb(telegram_received_cb)

        if device_updated_cb is not None:
            self.devices.register_device_updated_cb(device_updated_cb)


    def __del__(self):
        """Destructor. Cleaning up if this was not done before."""
        if self.started:
            try:
                task = asyncio.Task(self.stop())
                self.loop.run_until_complete(task)
            except RuntimeError as exp:
                self.logger.warning("Could not close loop, reason: %s", exp)


    @asyncio.coroutine
    def start(self,
              state_updater=False,
              daemon_mode=False,
              connection_config=ConnectionConfig()):
        """Start XKNX module. Connect to KNX/IP devices and start state updater."""
        self.knxip_interface = KNXIPInterface(self, connection_config=connection_config)
        yield from self.knxip_interface.start()
        yield from self.telegram_queue.start()

        if state_updater:
            from xknx.core import StateUpdater
            self.state_updater = StateUpdater(self)
            yield from self.state_updater.start()

        if daemon_mode:
            yield from self.loop_until_sigint()

        self.started = True

    @asyncio.coroutine
    def join(self):
        """Wait until all telegrams were processed."""
        yield from self.telegrams.join()


    def _stop_knxip_interface_if_exists(self):
        """Stop KNXIPInterface if initialized."""
        if self.knxip_interface is not None:
            yield from self.knxip_interface.stop()
            self.knxip_interface = None

    @asyncio.coroutine
    def stop(self):
        """Stop XKNX module."""
        yield from self.join()
        yield from self.telegram_queue.stop()
        yield from self._stop_knxip_interface_if_exists()
        self.started = False

    @asyncio.coroutine
    def loop_until_sigint(self):
        """Loop until Crtl-C was pressed."""
        def sigint_handler():
            """Callback for having Ctrl-C received."""
            self.sigint_received.set()
        self.loop.add_signal_handler(signal.SIGINT, sigint_handler)
        self.logger.warning('Press Ctrl+C to stop')
        yield from self.sigint_received.wait()



"""
Module for managing the climate within a room.

* It reads/listens to a temperature address from KNX bus.
* Manages and sends the desired setpoint to KNX bus.
"""
import time
import datetime
import asyncio
from enum import Enum
from xknx.knx import Address, DPTBinary, DPTArray, DPTTemperature
from xknx.exceptions import CouldNotParseTelegram
from .device import Device


class ClimateOperationMode(Enum):
    """Enum for the different KNX climate operation modes."""
    COMFORT = "Comfort"
    STANDBY = "Standby"
    NIGHT = "Night"
    PROTECTION = "Frost Protection"

class Climate(Device):
    """Class for managing the climate."""

    # pylint: disable=too-many-instance-attributes,invalid-name

    def __init__(self,
                 xknx,
                 name,
                 group_address_temperature=None,
                 group_address_target_temperature=None,
                 group_address_setpoint=None,
                 group_address_operation_mode=None,
                 group_address_operation_mode_protection=None,
                 group_address_operation_mode_night=None,
                 group_address_operation_mode_comfort=None,
                 device_updated_cb=None):
        """Initialize Climate class."""
        # pylint: disable=too-many-arguments
        Device.__init__(self, xknx, name, device_updated_cb)
        if isinstance(group_address_temperature, (str, int)):
            group_address_temperature = Address(group_address_temperature)
        if isinstance(group_address_target_temperature, (str, int)):
            group_address_target_temperature = Address(group_address_target_temperature)
        if isinstance(group_address_setpoint, (str, int)):
            group_address_setpoint = Address(group_address_setpoint)
        if isinstance(group_address_operation_mode, (str, int)):
            group_address_operation_mode = Address(group_address_operation_mode)
        if isinstance(group_address_operation_mode_protection, (str, int)):
            group_address_operation_mode_protection = Address(group_address_operation_mode_protection)
        if isinstance(group_address_operation_mode_night, (str, int)):
            group_address_operation_mode_night = Address(group_address_operation_mode_night)
        if isinstance(group_address_operation_mode_comfort, (str, int)):
            group_address_operation_mode_comfort = Address(group_address_operation_mode_comfort)

        self.group_address_temperature = group_address_temperature
        self.group_address_target_temperature = group_address_target_temperature
        self.group_address_setpoint = group_address_setpoint
        self.group_address_operation_mode = group_address_operation_mode
        self.group_address_operation_mode_protection = group_address_operation_mode_protection
        self.group_address_operation_mode_night = group_address_operation_mode_night
        self.group_address_operation_mode_comfort = group_address_operation_mode_comfort

        self.last_set = None
        self.temperature = None
        self.target_temperature = None
        self.setpoint = None
        self.operation_mode = ClimateOperationMode.STANDBY

        self.supports_temperature = \
            group_address_temperature is not None
        self.supports_target_temperature = \
            group_address_target_temperature is not None
        self.supports_setpoint = \
            group_address_setpoint is not None
        self.supports_operation_mode = \
            group_address_operation_mode is not None or \
            group_address_operation_mode_protection is not None or \
            group_address_operation_mode_night is not None or \
            group_address_operation_mode_comfort is not None

    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address_temperature = \
            config.get('group_address_temperature')
        group_address_target_temperature = \
            config.get('group_address_target_temperature')
        group_address_setpoint = \
            config.get('group_address_setpoint')
        group_address_operation_mode = \
            config.get('group_address_operation_mode')
        group_address_operation_mode_protection = \
            config.get('group_address_operation_mode_protection')
        group_address_operation_mode_night = \
            config.get('group_address_operation_mode_night')
        group_address_operation_mode_comfort = \
            config.get('group_address_operation_mode_comfort')
        return cls(xknx,
                   name,
                   group_address_temperature=group_address_temperature,
                   group_address_target_temperature=group_address_target_temperature,
                   group_address_setpoint=group_address_setpoint,
                   group_address_operation_mode=group_address_operation_mode,
                   group_address_operation_mode_protection=group_address_operation_mode_protection,
                   group_address_operation_mode_night=group_address_operation_mode_night,
                   group_address_operation_mode_comfort=group_address_operation_mode_comfort)

    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return self.group_address_temperature == group_address or \
               self.group_address_target_temperature == group_address or \
               self.group_address_setpoint == group_address or \
               self.group_address_operation_mode == group_address or \
               self.group_address_operation_mode_protection == group_address or \
               self.group_address_operation_mode_night == group_address or \
               self.group_address_operation_mode_comfort == group_address

    @asyncio.coroutine
    def _set_internal_setpoint(self, setpoint):
        """Set internal value of setpoint. Call hooks if setpoint was changed."""
        if setpoint != self.setpoint:
            self.setpoint = setpoint
            yield from self.after_update()

    @asyncio.coroutine
    def _set_internal_temperature(self, temperature):
        """Set internal value of temperature. Call hooks if setpoint was changed."""
        if temperature != self.temperature:
            self.temperature = temperature
            yield from self.after_update()

    @asyncio.coroutine
    def _set_internal_target_temperature(self, target_temperature):
        """Set internal value of target temperature. Call hooks if setpoint was changed."""
        if target_temperature != self.target_temperature:
            self.target_temperature = target_temperature
            yield from self.after_update()

    @asyncio.coroutine
    def _set_internal_operation_mode(self, operation_mode):
        """Set internal value of operatio nmode. Call hooks if setpoint was changed."""
        if operation_mode != self.operation_mode:
            self.operation_mode = operation_mode
            yield from self.after_update()

    @asyncio.coroutine
    def set_setpoint(self, setpoint):
        """Send setpoint to KNX bus."""
        if not self.supports_setpoint:
            return
        yield from self.send(
            self.group_address_setpoint,
            DPTArray(DPTTemperature().to_knx(setpoint)))
        yield from self._set_internal_setpoint(setpoint)

    @asyncio.coroutine
    def set_target_temperature(self, target_temperature):
        """Calculate setpoint-delta and setpoint and send it to  KNX bus."""
        setpoint_delta = self.setpoint - self.target_temperature
        new_setpoint = target_temperature + setpoint_delta
        yield from self.set_setpoint(new_setpoint)

    @asyncio.coroutine
    def set_operation_mode(self, operation_mode):
        """Set the operation mode of a thermostat. Send new operation_mode to BUS and update internal state."""
        if not self.supports_operation_mode:
            return
        if self.group_address_operation_mode is not None:
            knx_operation_mode = self._to_knx_operation_mode(operation_mode)
            yield from self.send(
                self.group_address_operation_mode,
                DPTArray(knx_operation_mode))
        if self.group_address_operation_mode_protection is not None:
            protection_mode = operation_mode == ClimateOperationMode.PROTECTION
            yield from self.send(
                self.group_address_operation_mode_protection,
                DPTBinary(protection_mode))
        if self.group_address_operation_mode_night is not None:
            night_mode = operation_mode == ClimateOperationMode.NIGHT
            yield from self.send(
                self.group_address_operation_mode_night,
                DPTBinary(night_mode))
        if self.group_address_operation_mode_comfort is not None:
            comfort_mode = operation_mode == ClimateOperationMode.COMFORT
            yield from self.send(
                self.group_address_operation_mode_comfort,
                DPTBinary(comfort_mode))
        yield from self._set_internal_operation_mode(operation_mode)

    def get_supported_operation_modes(self):
        """Returns all configured operation modes."""
        if not self.supports_operation_mode:
            return []
        if self.group_address_operation_mode is not None:
            return list(ClimateOperationMode)
        operation_modes = []
        if self.group_address_operation_mode_comfort:
            operation_modes.append(ClimateOperationMode.COMFORT)
        operation_modes.append(ClimateOperationMode.STANDBY)
        if self.group_address_operation_mode_night:
            operation_modes.append(ClimateOperationMode.NIGHT)
        if self.group_address_operation_mode_protection:
            operation_modes.append(ClimateOperationMode.PROTECTION)
        return operation_modes

    @asyncio.coroutine
    def process(self, telegram):
        """Process incoming telegram."""
        if telegram.group_address == self.group_address_temperature and \
                self.supports_temperature:
            yield from self._process_temperature(telegram)
        if telegram.group_address == self.group_address_target_temperature and \
                self.supports_target_temperature:
            yield from self._process_target_temperature(telegram)
        elif telegram.group_address == self.group_address_setpoint and \
                self.supports_setpoint:
            yield from self._process_setpoint(telegram)
        elif telegram.group_address == self.group_address_operation_mode and \
                self.supports_operation_mode:
            yield from self._process_operation_mode(telegram)
        # Note: telegrams setting splitted up operation modes are not yet implemented

    @asyncio.coroutine
    def _process_temperature(self, telegram):
        """Process incoming telegram for temperature."""
        if not isinstance(telegram.payload, DPTArray) \
                or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()
        temperature = DPTTemperature().from_knx(
            (telegram.payload.value[0],
             telegram.payload.value[1]))
        self.last_set = time.time()
        yield from self._set_internal_temperature(temperature)

    @asyncio.coroutine
    def _process_target_temperature(self, telegram):
        """Process incoming telegram for target temperature."""
        if not isinstance(telegram.payload, DPTArray) \
                or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()
        target_temperature = DPTTemperature().from_knx(
            (telegram.payload.value[0],
             telegram.payload.value[1]))
        self.last_set = time.time()
        yield from self._set_internal_target_temperature(target_temperature)

    @asyncio.coroutine
    def _process_setpoint(self, telegram):
        """Process incoming telegram for setpoint."""
        if not isinstance(telegram.payload, DPTArray) \
                or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()
        setpoint = DPTTemperature().from_knx(
            (telegram.payload.value[0],
             telegram.payload.value[1]))
        yield from self._set_internal_setpoint(setpoint)

    @asyncio.coroutine
    def _process_operation_mode(self, telegram):
        """Process incoming telegram for operation mode."""
        if not isinstance(telegram.payload, DPTArray) \
            or len(telegram.payload.value) != 1:
            raise CouldNotParseTelegram()
        knx_operation_mode = telegram.payload.value[0]
        operation_mode = self._from_knx_operation_mode(knx_operation_mode)
        yield from self._set_internal_operation_mode(operation_mode)

    def state_addresses(self):
        """Return group addresses which should be requested to sync state."""
        state_addresses = []
        if self.supports_temperature:
            state_addresses.append(self.group_address_temperature)
        if self.supports_target_temperature:
            state_addresses.append(self.group_address_target_temperature)
        if self.supports_setpoint:
            state_addresses.append(self.group_address_setpoint)
        if self.supports_operation_mode:
            if self.group_address_operation_mode:
                state_addresses.append(self.group_address_operation_mode)
            # Note: telegrams setting splitted up operation modes are not yet implemented
        return state_addresses

    @staticmethod
    def _to_knx_operation_mode(operation_mode):
        """Return the KNX value for the operation mode."""
        if operation_mode == ClimateOperationMode.COMFORT:
            return 1
        elif operation_mode == ClimateOperationMode.STANDBY:
            return 2
        elif operation_mode == ClimateOperationMode.NIGHT:
            return 3
        elif operation_mode == ClimateOperationMode.PROTECTION:
            return 4

    @staticmethod
    def _from_knx_operation_mode(knx_operation_mode):
        """Return the enum value for a KNX value of an operation mode."""
        if knx_operation_mode == 1:
            return ClimateOperationMode.COMFORT
        elif knx_operation_mode == 2:
            return ClimateOperationMode.STANDBY
        elif knx_operation_mode == 3:
            return ClimateOperationMode.NIGHT
        elif knx_operation_mode == 4:
            return ClimateOperationMode.PROTECTION

    def __str__(self):
        """Return object as readable string."""
        last_set_formatted = \
                datetime.datetime.fromtimestamp(
                    self.last_set).strftime('%Y-%m-%d %H:%M:%S') \
                if self.last_set else None
        return '<Climate name="{0}" ' \
               'group_address_temperature="{1}"  ' \
               'group_address_target_temperature="{2}"  ' \
               'group_address_setpoint="{3}" ' \
               'group_address_operation_mode="{4}" ' \
               'temperature="{5}" ' \
               'last_set="{6}" />' \
               .format(self.name,
                       self.group_address_temperature,
                       self.group_address_target_temperature,
                       self.group_address_setpoint,
                       self.group_address_operation_mode,
                       self.temperature,
                       last_set_formatted)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



""" Implementation of Basic KNX 2-Byte """

from xknx.exceptions import ConversionError
from .dpt import DPTBase

class DPT2Byte(DPTBase):
    """
    Abstraction for KNX 2 Byte "2-octet unsigned counter value"
    DPT 7.001
    """

    value_min = 0
    value_max = 65535
    unit = "pulses"
    resolution = 1

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 2)
        return (raw[0] * 256) + raw[1]

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if not cls._test_boundaries(value):
            raise ConversionError(value)
        return value >> 8, value  & 0xff

    @classmethod
    def _test_boundaries(cls, value):
        """Test if value is within defined range for this object."""
        return value >= cls.value_min and \
               value <= cls.value_max

class DPTUElCurrentmA(DPT2Byte):
    """
    Abstraction for KNX 2 Byte DPTUElCurrentmA
    DPT 7.012
    """

    value_min = 0
    value_max = 65535
    unit = "mA"
    resolution = 1



"""
Module for serialization and deserialization of KNX HPAI (Host Protocol Address Information) information.

A HPAI contains an IP address and a port.
"""

from xknx.exceptions import ConversionError, CouldNotParseKNXIP

class HPAI():
    """Class for Module for Serialization and Deserialization."""

    LENGTH = 0x08
    TYPE_UDP = 0x01

    def __init__(self, ip_addr='0.0.0.0', port=0):
        self.ip_addr = ip_addr
        self.port = port

    def from_knx(self, raw):
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < HPAI.LENGTH:
            raise CouldNotParseKNXIP("wrong HPAI length")
        if raw[0] != HPAI.LENGTH:
            raise CouldNotParseKNXIP("wrong HPAI length")
        if raw[1] != HPAI.TYPE_UDP:
            raise CouldNotParseKNXIP("wrong HPAI type")

        self.ip_addr = "{0}.{1}.{2}.{3}".format(
            raw[2], raw[3], raw[4], raw[5])
        self.port = raw[6]*256 + raw[7]

        return HPAI.LENGTH

    def to_knx(self):
        """Serialize to KNX/IP raw data."""
        def ip_addr_to_bytes(ip_addr):
            """Serialize ip address to byte array."""
            if not isinstance(ip_addr, str):
                raise ConversionError("ip_addr is not a string")
            for i in ip_addr.split("."):
                yield int(i)
        data = []
        data.append(HPAI.LENGTH)
        data.append(HPAI.TYPE_UDP)
        data.extend(ip_addr_to_bytes(self.ip_addr))
        data.append((self.port >> 8) & 255)
        data.append(self.port & 255)
        return data

    def __str__(self):
        """Return object as readable string."""
        return '<HPAI {0}:{1} />'.format(self.ip_addr, self.port)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__




""" Implementation of Basic KNX Floats """

from xknx.exceptions import ConversionError
from .dpt import DPTBase

class DPTFloat(DPTBase):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers
    DPT 9.xxx
    """

    value_min = -671088.64
    value_max = 670760.96
    unit = ""
    resolution = 1

    @classmethod
    def from_knx(cls, raw):
        """Parse/deserialize from KNX/IP raw data."""
        cls.test_bytesarray(raw, 2)
        data = (raw[0] * 256) + raw[1]
        exponent = (data >> 11) & 0x0f
        significand = data & 0x7ff
        sign = data >> 15

        if sign == 1:
            significand = significand - 2048

        value = float(significand << exponent) / 100

        if not cls._test_boundaries(value):
            raise ConversionError(value)

        return value

    @classmethod
    def to_knx(cls, value):
        """Serialize to KNX/IP raw data."""
        if not isinstance(value, (int, float)):
            raise ConversionError(value)
        if not cls._test_boundaries(value):
            raise ConversionError(value)
        sign = 1 if value < 0 else 0
        def calc_exponent(value, sign):
            """Return float exponent."""
            exponent = 0
            significand = abs(int(value * 100))

            while significand < -2048 or significand > 2048:
                exponent += 1
                significand >>= 1

            if sign:
                significand ^= 0x7ff # invert
                significand += 1     # and add 1

            return exponent, significand

        exponent, significand = calc_exponent(value, sign)

        return (sign << 7) | (exponent << 3) | (significand >> 8), \
               significand & 0xff

    @classmethod
    def _test_boundaries(cls, value):
        """Test if value is within defined range for this object."""
        return value >= cls.value_min and \
               value <= cls.value_max

class DPTTemperature(DPTFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers
    DPT 9.001
    """

    value_min = -273
    value_max = 670760
    unit = "C"
    resolution = 1

class DPTLux(DPTFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers
    DPT 9.004
    """

    value_min = 0
    value_max = 670760
    unit = "Lux"
    resolution = 1

class DPTWsp(DPTFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers
    DPT 9.005
    """

    value_min = 0
    value_max = 670760
    unit = "m/s"
    resolution = 1

class DPTHumidity(DPTFloat):
    """
    Abstraction for KNX 2 Octet Floating Point Numbers
    DPT 9.007
    """

    value_min = 0
    value_max = 670760
    unit = "%"
    resolution = 1



"""
AddressFilter provides a mechanism for filtering
KNX addresses with patterns.

Patterns can be

    for level3 KNX addresses:

        AddressFilter("1/*/2-5")
        AddressFilter("1/1-3,4,5/*")
        AddressFilter("1/2/-10)

    for level2 KNX addresses:

        AddressFilter("*/2-5")
        AddressFilter("1-3,4,5/*")
        AddressFilter("2/-10")

    for free format KNX addresses:

        AddressFilter("2-5")
        AddressFilter("1-3,4,5")
        AddressFilter("-10")
"""
from xknx.exceptions import ConversionError
from .address import Address

class AddressFilter:
    """Class for filtering Addresses according to patterns."""

    # pylint: disable=too-few-public-methods

    def __init__(self, pattern):
        """Initialization of AddressFilter class."""
        self.level_filters = []
        self._parse_pattern(pattern)

    def _parse_pattern(self, pattern):
        for part in pattern.split("/"):
            if len(part) == 0:
                raise ConversionError("Every part of pattern must be a string.")
            self.level_filters.append(AddressFilter.LevelFilter(part))
        if len(self.level_filters) > 3:
            raise ConversionError("Too many parts within pattern.")

    def match(self, address):
        """Test if provided address matches Addressfilter."""
        if isinstance(address, str):
            address = Address(address)
        if len(self.level_filters) == 3:
            return self._match_level3(address)
        elif len(self.level_filters) == 2:
            return self._match_level2(address)
        elif len(self.level_filters) == 1:
            return self._match_free(address)

    def _match_level3(self, address):
        return (
            self.level_filters[0].match(address.get_level3_1())
            and
            self.level_filters[1].match(address.get_level3_2())
            and
            self.level_filters[2].match(address.get_level3_3()))

    def _match_level2(self, address):
        return (
            self.level_filters[0].match(address.get_level2_1())
            and
            self.level_filters[1].match(address.get_level2_2()))

    def _match_free(self, address):
        return (
            self.level_filters[0].match(address.get_free()))


    class Range:
        """Class for filtering patterns like "8", "*", "8-10"."""
        def __init__(self, pattern):
            self.range_from = None
            self.range_to = None
            self._parse_pattern(pattern)

        def _parse_pattern(self, pattern):
            if pattern == "*":
                self._init_wildcard()
            elif pattern.isdigit():
                self._init_digit(pattern)
            elif "-" in pattern:
                self._init_range(pattern)
            self.range_to = self._adjust_range(self.range_to)
            self.range_from = self._adjust_range(self.range_from)
            self._flip_range_if_necessary()

        def _init_wildcard(self):
            self.range_from = 0
            self.range_to = Address.MAX_FREE

        def _init_digit(self, pattern):
            digit = int(pattern)
            self.range_from = digit
            self.range_to = digit

        def _init_range(self, pattern):
            (range_from, range_to) = pattern.split("-")
            self.range_from = int(range_from) \
                if len(range_from) > 0 else 0
            self.range_to = int(range_to) \
                if len(range_to) > 0 else Address.MAX_FREE

        @staticmethod
        def _adjust_range(digit):
            if digit > Address.MAX_FREE:
                return Address.MAX_FREE
            if digit < 0:
                return 0
            return digit

        def _flip_range_if_necessary(self):
            if self.range_from > self.range_to:
                self.range_to, self.range_from =\
                    self.range_from, self.range_to

        def get_range(self):
            """Return the range (from,to) of this pattern."""
            return self.range_from, self.range_to

        def match(self, digit):
            """Return if given digit is within range of pattern."""
            return \
                digit >= self.range_from and \
                digit <= self.range_to


    class LevelFilter:
        """Class for filtering patterns like "8,11-14,20"."""
        # pylint: disable=too-few-public-methods
        def __init__(self, pattern):
            self.ranges = []
            self._parse_pattern(pattern)

        def _parse_pattern(self, pattern):
            for part in pattern.split(","):
                if len(part) == 0:
                    raise ConversionError("Every part of LevelFilter must be a string")
                self.ranges.append(AddressFilter.Range(part))

        def match(self, digit):
            """Return if given digit is within range of pattern."""
            for _range in self.ranges:
                if _range.match(digit):
                    return True
            return False



"""
Module for XKXN Exceptions.
"""

class XKNXException(Exception):
    """Default XKNX Exception."""

    pass


class CouldNotParseTelegram(XKNXException):
    """Could not parse telegram error."""

    def __init__(self, description=""):
        """Initialize CouldNotParseTelegram class."""
        super(CouldNotParseTelegram, self).__init__("Could not parse Telegram")
        self.description = description

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseTelegram description="{0}" />' \
            .format(self.description)


class CouldNotParseKNXIP(XKNXException):
    """Exception falss for worong KNXIP data."""

    def __init__(self, description=""):
        """Initialize CouldNotParseTelegram class."""
        super(CouldNotParseKNXIP, self).__init__("Could not parse KNXIP")
        self.description = description

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseKNXIP description="{0}" />' \
            .format(self.description)


class ConversionError(XKNXException):
    """Exception class for error while converting one type to another."""

    def __init__(self, i):
        """Initialize ConversionError class."""
        super(ConversionError, self).__init__("Conversion Error")
        self.i = i

    def __str__(self):
        """Return object as readable string."""
        return '<ConversionError input="{0}" />'.format(self.i)


class CouldNotParseAddress(XKNXException):
    """Exception class for wrong address format."""

    def __init__(self, address=None):
        """Initialize CouldNotParseAddress class."""
        super(CouldNotParseAddress, self).__init__("Could not parse address")
        self.address = address

    def __str__(self):
        """Return object as readable string."""
        return '<CouldNotParseAddress address="{0}" />'.format(self.address)



"""
Module for broadcasting time to KNX bus.

Time is a virtual/pseudo device, using the infrastructure for
beeing configured via xknx.yaml and synchronized periodically
by StateUpdate.
"""

import asyncio
from xknx.knx import Address, DPTArray, DPTTime
from .device import Device


class Time(Device):
    """Class for virtual time device."""

    def __init__(self,
                 xknx,
                 name,
                 group_address=None,
                 device_updated_cb=None):
        """Initialize Time class."""
        Device.__init__(self, xknx, name, device_updated_cb)

        if isinstance(group_address, (str, int)):
            group_address = Address(group_address)

        self.group_address = group_address


    @classmethod
    def from_config(cls, xknx, name, config):
        """Initialize object from configuration structure."""
        group_address = \
            config.get('group_address')

        return cls(xknx,
                   name,
                   group_address=group_address)


    def has_group_address(self, group_address):
        """Test if device has given group address."""
        return self.group_address == group_address


    @asyncio.coroutine
    def broadcast_time(self):
        """Broadcast time to KNX bus."""
        yield from self.send(
            self.group_address,
            DPTArray(DPTTime.current_time_as_knx()))

    @asyncio.coroutine
    def sync(self, wait_for_result=True):
        """Read state of device from KNX bus. Used here to broadcast time to KNX bus."""
        yield from self.broadcast_time()

    def __str__(self):
        """Return object as readable string."""
        return '<Time name="{0}" group_address="{1}" />' \
            .format(self.name, self.group_address)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__
